import hashlib
import json
from typing import Any, Dict, List
from rightchain.domain.models.file import File
from rightchain.domain.models.repo_status import RepositoryStatus
from rightchain.infra.copyright_store_service import CopyrightStoreService
from rightchain.infra.file_lister import FileListerService
import secrets

from rightchain.infra.git_service import GitService


class IndexingService:
    def __init__(
        self,
        fileListerService: FileListerService,
        copyrightStoreService: CopyrightStoreService,
        gitService: GitService,
    ) -> None:
        self.fileListerService = fileListerService
        self.copyrightStoreService = copyrightStoreService
        self.gitService = gitService

    def fileSha256(self, filename: str) -> str:
        with open(filename, "rb") as f:
            h = hashlib.sha256()
            h.update(f.read())
            return h.hexdigest().lower()

    def stringSha256(self, s: str) -> str:
        h = hashlib.sha256()
        h.update(s.encode("utf8"))
        return h.hexdigest().lower()

    def generateSalt(self, nbytes: int) -> str:
        return secrets.token_hex(nbytes).lower()

    def GetRepositoryStatus(self) -> RepositoryStatus:
        filenames = self.fileListerService.list_files_with_gitignore_and_rightignore()

        files: List[File] = []

        for filename in filenames:
            salt: str = self.generateSalt(30)
            hash = self.stringSha256(self.fileSha256(filename) + salt)
            files.append(File(filename, salt, hash))

        copyrightInfo = self.fileListerService.TryReadCopyrightInfo()
        commit = self.gitService.GetPreviousCommitHash()

        return RepositoryStatus(files, copyrightInfo, commit)

    def GenerateSaltInfo(self, status: RepositoryStatus):
        return {
            "warning": "This file should keep secret. When a file is needed, only the salt of it should be exposed",
            "hint": "for a file, sha256(sha256(content_of_file)+salt) = sha256_in_index_json",
            "data": [{"filename": p.filename, "salt": p.salt} for p in status.files],
        }

    def GenerateIndexInfo(self, status: RepositoryStatus):
        return {
            "previousCommit": status.commit,
            "copyrightInfo": status.copyrightInfo,
            "data": [p.hash for p in status.files],
        }

    def GetSha256OfIndexFile(self) -> str:
        return self.fileSha256(self.copyrightStoreService.index_file)




from typing import List, Optional
from rightchain.domain.models.file import File


class RepositoryStatus:
    def __init__(self, files: List[File], copyrightInfo: Optional[str], commit:str) -> None:
        self.files = list(files)
        self.copyrightInfo = copyrightInfo
        self.commit = commit



import json
import os
from typing import Any

from rightchain.domain.models.index_record import IndexRecord
from rightchain.domain.services.indexing import IndexingService
from rightchain.infra.client import RightClient
from rightchain.infra.copyright_store_service import CopyrightStoreService


class MyController:
    def __init__(
        self,
        indexingService: IndexingService,
        copyrightStoreService: CopyrightStoreService,
        rightClient: RightClient,
    ) -> None:
        self.indexingService = indexingService
        self.copyrightStoreService = copyrightStoreService
        self.rightClient = rightClient

    def CreateIndex(self) -> None:
        repoStatus = self.indexingService.GetRepositoryStatus()
        saltInfo = self.indexingService.GenerateSaltInfo(repoStatus)
        indexInfo = self.indexingService.GenerateIndexInfo(repoStatus)

        self.copyrightStoreService.WriteSaltFile(saltInfo)
        self.copyrightStoreService.WriteIndexFile(indexInfo)

        print(f"CreateIndex ok")

    def PushIndex(self) -> None:
        indexFileHash = self.indexingService.GetSha256OfIndexFile()

        commit = self.copyrightStoreService.ReadIndexFile()["previousCommit"]
        assert isinstance(commit, str)

        token = self.rightClient.outofbox_create_record(indexFileHash)

        # responsibility transfer
        self.copyrightStoreService.SaveIndexRecord(IndexRecord(commit, token, None))

        print(f"Pushed and create waiting file: {commit}")

    def Fetch(self) -> None:
        def is_packaged(record: dict):
            return record["transactionId"] is not None

        for indexRecord in self.copyrightStoreService.GetWaitingIndexRecords():

            recordInfo = self.rightClient.outofbox_get_record(indexRecord.token)

            if is_packaged(recordInfo):
                indexRecord.UpdateInfo(recordInfo)
                assert not indexRecord.IsWaiting

                self.copyrightStoreService.SaveIndexRecord(indexRecord)
                print("packaged:", indexRecord.commit)
            else:
                print("still not packaged:", indexRecord.commit)



from typing import Any


class IndexRecord:
    def __init__(self, commit: str, token: "str|None", recordInfo: Any) -> None:
        self.commit = commit
        self.token = token
        self.recordInfo: Any = recordInfo

    @property
    def IsWaiting(self) -> bool:
        return self.token is not None

    def UpdateInfo(self, recordInfo) -> None:
        assert self.IsWaiting
        assert recordInfo is not None
        assert isinstance(recordInfo, dict)

        self.token = None
        self.recordInfo = recordInfo




from typing import List
import pathspec
import os

from rightchain.infra.copyright_store_service import CopyrightStoreService


class FileListerService:
    def __init__(self, copyrightStoreService: CopyrightStoreService) -> None:
        self.copyrightStoreService = copyrightStoreService
        pass

    def list_all_files_in_workdir(self) -> List[str]:
        all_files: List[str] = []

        for dirpath, dirs, files in os.walk("."):
            for name in files:
                filename = os.path.join(dirpath, name)
                filename = filename.replace("\\", "/")
                all_files.append(filename)

        return all_files

    def list_included_file_in_workdir(self, exclude_spec_lines: List[str]):
        spec = pathspec.PathSpec.from_lines(
            pathspec.patterns.GitWildMatchPattern, exclude_spec_lines
        )

        all_files = self.list_all_files_in_workdir()
        included: List[str] = []

        for file in all_files:
            if not spec.match_file(file):
                included.append(file)

        return included

    def list_files_with_gitignore_and_rightignore(self):
        exclude_spec_lines = []
        exclude_spec_lines.append(".git")
        exclude_spec_lines.append(self.copyrightStoreService.right_dir)

        for ignorefile in [".gitignore", ".rightignore"]:
            if os.path.exists(ignorefile):
                print(f"using ignore file: {ignorefile}")

                with open(ignorefile, "r", encoding="utf8") as f:
                    lines = f.readlines()

                exclude_spec_lines += lines

        files = self.list_included_file_in_workdir(exclude_spec_lines)
        return files

    def TryReadCopyrightInfo(self) -> "str|None":
        copyright_info = "copyright-info"
        if os.path.exists(copyright_info):
            print(f"using copyright info: {copyright_info}")

            with open(copyright_info, "r", encoding="utf8") as f:
                info = f.read()
                return info
        else:
            return None


# files = list_files_with_gitignore_and_rightignore()
# print(len(files))

# for file in files:
#     print(file)



import subprocess


class GitService:
    def __init__(self) -> None:
        pass

    def GetPreviousCommitHash(self) -> str:
        cmd = "git log --pretty=format:%H -1"
        outp = subprocess.check_output(cmd)
        return outp.decode("utf8")




import json
import os
import pathlib
from typing import Any, Dict, List, Tuple

from rightchain.domain.models.index_record import IndexRecord


class CopyrightStoreService:
    def __init__(self) -> None:
        self.right_dir = "copyrightstore"

        self.wait_dir = f"{self.right_dir}/waiting"
        self.packaged_dir = f"{self.right_dir}/packaged"
        self.index_file = f"{self.right_dir}/index.json"
        self.salt_file = f"{self.right_dir}/salt.json"

        self.ensure_all_dir_created()

    def create_dir(self, path: str):
        fileinfo = pathlib.Path(path)
        if fileinfo.exists():
            assert fileinfo.is_dir()
        else:
            fileinfo.mkdir()

    def writeJsonToFile(self, filename: str, obj: Any) -> None:
        content = json.dumps(obj, ensure_ascii=True, indent=4)
        with open(filename, "w", encoding="utf8") as f:
            f.write(content)

    def readJsonFromFile(self, filename: str) -> Any:
        with open(filename, "r", encoding="utf8") as f:
            return json.load(f)

    def ensure_all_dir_created(self):
        self.create_dir(self.right_dir)
        self.create_dir(self.wait_dir)
        self.create_dir(self.packaged_dir)

    def WriteIndexFile(self, obj: Dict) -> None:
        self.writeJsonToFile(self.index_file, obj)

    def WriteSaltFile(self, obj: Dict) -> None:
        self.writeJsonToFile(self.salt_file, obj)

    def ReadIndexFile(self) -> Dict[str, Any]:
        return self.readJsonFromFile(self.index_file)

    def SaveIndexRecord(self, record: IndexRecord) -> None:
        waitingPath = os.path.join(self.wait_dir, record.commit)
        packagedPath = os.path.join(self.packaged_dir, record.commit)

        if record.IsWaiting:
            "save to waiting"
            self.writeJsonToFile(waitingPath, {"token": record.token})

        else:
            "remove waiting"
            if os.path.exists(waitingPath):
                os.remove(waitingPath)

            "save to packaged"
            self.writeJsonToFile(packagedPath, record.recordInfo)

    def GetWaitingIndexRecords(self) -> List[IndexRecord]:
        items: List[IndexRecord] = []
        for name in os.listdir(self.wait_dir):
            commit_hash = name
            filename = os.path.join(self.wait_dir, name)
            json_info = self.readJsonFromFile(filename)

            token = json_info["token"]

            items.append(IndexRecord(commit_hash, token, None))

        return items




import argparse

from rightchain.application.my_controller import MyController
from rightchain.domain.services.indexing import IndexingService
from rightchain.infra.client import RightClient
from rightchain.infra.copyright_store_service import CopyrightStoreService
from rightchain.infra.file_lister import FileListerService
from rightchain.infra.git_service import GitService

if __name__ == "__main__":
    gitService = GitService()
    rightClient = RightClient()
    copyrightStoreService = CopyrightStoreService()
    fileListerService = FileListerService(copyrightStoreService)
    indexingService = IndexingService(
        fileListerService, copyrightStoreService, gitService
    )
    controller = MyController(indexingService, copyrightStoreService, rightClient)

    parser = argparse.ArgumentParser(
        prog="rccc", description="Commandline tool for copyright on blockchain"
    )

    # parser.add_argument('command', choices=[
    #                     'create-index', 'push-index', 'index', 'fetch', 'test-login'], help="action to perform")

    parser.add_argument(
        "command",
        choices=["create-index", "push-index", "index", "fetch"],
        help="action to perform",
    )

    args = parser.parse_args()

    command = args.command

    if command == "create-index":
        controller.CreateIndex()
    elif command == "push-index":
        controller.PushIndex()
        print("Pushed index to server.")
    elif command == "index":
        controller.CreateIndex()
        controller.PushIndex()
    elif command == "fetch":
        controller.Fetch()
        print("Fetch from server complete")
    else:
        raise Exception("Not reached")
